#####################################################################################################################
# Licensed Materials - Property of IBM
# Rational ClearQuest
# (C) Copyright IBM Corp. 2009-2011.  All Rights Reserved.
# US Government Users Restricted Rights -
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp
######################################################################################################################
# 
#This script utilized the ClearQuest COM API and the ClearCase cleartool commands.
# 
# Variable used in this script are as followings:
#	user		The login id of the ClearQuest user
#	pw		The password of the ClearQuest user
#	dbname		The schema repository connection to the database
#	dbset		The user database connection
#	pvob		The name of the project vob
#	baseline	The name of the baseline
#       projectid       The ID of the project in ClearQuest.  For example, ALM00000045
#       ucmstream       The name of the stream used in ClearCase
#	$stream_used	The name of the stream based on the pvob and baseline
#	$num_records	The number of records found in ClearQuest database
#	main		The main subroutine of the script. Added for error handling purposes.
#	get_stream	The subrountine to find the stream value based on pvob and baseline
#	get_activities	The subrountine given two baselines find the difference between them and get all the activities
#	find_contribs	The subrountine finds the contributing activites based on a set of activities
#	get_status	The subrountine finds the label status of a baseline
#	get_seed	The subrountine finds the foundation (seed) baseline based on the stream and pvob
#	get_project	The subrountine finds the project for the baseline
#	get_promotionlevel	The subrountine finds the promotion level for the baseline
#	get_components 	The subrountine finds the components of a given baseline
#	get_modcomponents	The subrountine finds the modifiable components of a baseline
#	found_baselines 	The subrountine iterates through all the foundation baselines to find the ones with modifiable components
#       total_activities  	The subrountine iterates over each foundation baseline to find the total activities
#   	create_new_record	The subroutine that creates and commits records to CQ.
#	logfile			The path where the log file will be sent
#	test			This is a flag that can be set for testing file inputs
#	help			This is a flag for the help
#	EDIT_BUILD_ACTION	This variable will set a session variable for ClearQuest access.
#
#######################################################################################################################
#!ratlperl -l -w
# Changed above line to read ratlperl from cqperl thl1
#TODO fix globals
my $VERSION = "7.1.0.3";
my $INTERNALVERSION = "0.7";
use Config;
my $perl;
if ($Config{osname} =~ /win32/i) {
	$perl = 'ratlperl';
} else {
	$perl = 'cqperl';
}
use strict;
use vars qw($Bin);
use FindBin qw($Bin);
use lib "$Bin";
use Getopt::Long;
use File::Basename;
use Carp;
use FileHandle;
use CQPerlExt;
use Env;
use IO::File;
use LWP::Simple;
use POSIX;
my %seen;
my @union =();
my @isect=();
my %union =();
my %isect =();
my %count = ();
my $result;
my $stream_used;
my $oldbaseline;
my $baseline_status;
my $promotion_level;
my $mod_comps;
my @foundbaseline;
my $foundbaseline;
my $unique;
my @list;
my ($entity);
my $userLogin ; #thl4 
my $oldPVOB ; 
my $project;
my $slashvob;
my $noslashvob;

#globals added when main sub split out
my $resultSet;
my $stream_used;
my %opts;


use subs qw(main get_stream get_activities find_contribs get_status get_seed get_project get_promotionlevel get_components get_modcomponents found_baselines total_activities create_new_record);

# Unbuffer STDERR and STDOUT
select STDERR;  
$| = 1;
select STDOUT;
$| = 1;

#Define Logger parameters
use Logger qw(CONSOLE_ON CONSOLE_OFF CONSOLE_TERSE);
my $logger = Logger->new(); # the logger element

# Create the utility object
use UtilLib;
my $ut = UtilLib->new();
my $log_time = $ut->ut_file_name_date_time_stamp();

my $progname = basename $0;
my $result = main();

if ($result) {
    exit(0); # DOS Success Code
} else {
    exit(1); # DOS Error Code
}

sub main {
####
# Define the input parameters
my @opts = qw(
    user=s		
    pw=s		
    dbname=s		
    dbset=s		
    pvob=s		
    baseline=s
    projectid=s
    logfile=s
    ucmstream=s
    test	
    help
    activity=s@
    from_file=s
    diffbl!
    create_composed_of_baselines!
    create_composed_of_baselines_level=i
    add_composed_of_baseline=s@
    add_pvob_or_location=s@
    component=s
);

my %opts = (
    user => undef,
    pw => undef,
    dbname => undef,
    dbset => undef,
    pvob => undef,
    baseline => undef,
    projectid => undef,
    logfile => undef,
    ucmstream => undef,      # Needed for nonUCM
    from_file => undef,
    activity => [],
    diffbl => 1,  
    create_composed_of_baselines => 1,
    create_composed_of_baselines_level => 1, #How many levels deep of add_composed_of_baseline baselines to create.  
    add_composed_of_baseline => [], # used to pass members instead of deriving from baselines.  NonUCM usage,  need to support bl@\pvob format
    add_pvob_or_location => [], # use to hold the corresponding "location" for each member baseline
    component => undef, # use to hold component of baseline.  It is not expected the user would pass this option in.
);

my @required_opts = qw ( user pw dbname dbset pvob baseline projectid);

####
# Get the input parameters
unless (GetOptions(\%opts, @opts)) {
 	$logger->log_print_message("R}hsւ̈p@:");	
      	my $retval = &Usage;
      	return 0;

}

#Define the LogFile
if ($opts{logfile}) {
	my $retval;
	my $logdir = dirname($opts{logfile});
	if ( ! -e $logdir) {
		$retval = $ut->ut_create_directory_tree($logdir);
		if ( ! $retval) {
            		$logger->log_print_error(__LINE__ . "O t@C fBNg $logdir 쐬ł܂ł");
            		$logger->log_print_error(__LINE__ . "XNvg̎śAȂŌp܂B");
        	}
	}
	$retval = $logger->log_set_file_on($opts{logfile});
	if ( ! $retval) {
        	$logger->log_print_error(__LINE__ . "O t@C $opts{logfile} J܂łB");
        	$logger->log_print_error(__LINE__ . "XNvg̎śAȂŌp܂B");
   	 }
	$logger->log_print_message("MO $log_time ɊJn܂B"); 
}	 

# Display the help usage
if ($opts{help}) {
	&Usage();
	return 1;
}
# Display test inputs to log file 
if ($opts{test}){
	$logger->log_print_message("`ꂽ[U[:");
	$logger->log_print_message("`ꂽ[U[ OC $opts{user} łB");
        $logger->log_print_message("`ꂽf[^x[X $opts{dbname} łB");
	$logger->log_print_message("`ꂽ dbset  $opts{dbset} łB");
	$logger->log_print_message("`ꂽvWFNg VOB  $opts{pvob} łB");
	$logger->log_print_message("`ꂽx[XC $opts{baseline} łB");	     
        $logger->log_print_message("`ꂽvWFNg ID  $opts{projectid} łB");
        $logger->log_print_message(" `ꂽ UCM Xg[ $opts{ucmstream} łB");
	if ($opts{logfile}){ $logger->log_print_message("`ꂽO t@C $opts{logfile} łB");}
	return 1;
}
####
# End program if any of the required parameters are not defined
{my  $message = "";
	for my $req (@required_opts )  {
		$message .= "K{wIvV -$req  ̌\n" if ! defined $opts{$req};
	}
	if ($message ) {
		$logger->log_print_message("KvȃR}hsIvVĂ܂AR}hsւ̈p@:");
		$logger->log_print_message("$message");
		my $retval = &Usage();
		return 0;
	}
}
;

if (defined $opts{from_file} ){
	 my $fh = new FileHandle "<$opts{from_file}";
	 if (defined $fh) {
		 my $err = 0;
		 while ( my $line = $fh->getline() ) {
			 chomp $line;
			 # Allow #comment in file, strip them off.
			 $line =~ s/#.*//;  
			 next if $line =~ /^\s*$/; #ignore blank lines.
			 #Match the activity in CQ format at the beginning of each line
			 #For example "ALM0001234"  "RATLC00001234"
			 if ( $line =~ /([a-zA-Z]+\d+)/ ) {
				 push @{$opts{activity}},$1;
			 }
			 else {
				 warn "$progname : t@C $opts{from_file} ̖ȃANeBreB: $line";
				 $err++;
			 }
		 }
		 die "G[̂߂ɑ݂Ă܂" if $err;
	 }
	 else {
		warn "$progname : -from_file '$opts{from_file}' : t@CȂAǂݎ邱Ƃł܂";
		my $retval = &Usage();
		exit 1;
	}
}
#### Start a CQ session
my $session = CQPerlExt::CQSession_Build();
# Set session variable 
#### Logon to the specified database

eval {$session->UserLogon("$opts{user}", "$opts{pw}", "$opts{dbname}", "$opts{dbset}")};
if ($@) {
$logger->log_print_error(__LINE__ . "$opts{dbname} ւ̃OCɎs܂: $@"); 
return 0;
}
$userLogin = $session->GetUserLoginName() ; #thl4
$logger->log_print_success("$userLogin ɂ $opts{dbname} ւ̃OC܂B"); #thl4

#### Set up vobs
##Passing in the vob to CQ requires no slashes on the vob, passing in to CC requires a slash on the vob
sub slash_noslash {
	my $pvob = shift;
	my $slashvob;
	my $noslashvob;
	if ($pvob !~ /[\/\\]/) {
	  $slashvob = '\\' .$pvob;
	  $noslashvob = $pvob;
	}
	else{
	  $slashvob = $pvob;
	  $noslashvob = $pvob ;
	  $noslashvob=~ s/\\//g; 
	}
	return ($slashvob, $noslashvob);
}
($slashvob, $noslashvob ) = &slash_noslash($opts{pvob});


#### Determine the stream used

if(!defined $opts{ucmstream}){
	$stream_used = get_stream( pvob => $slashvob, baseline => $opts{baseline} );
	if (! $stream_used){
		$logger->log_print_error(__LINE__ . "gpĂXg[𔻕ʂł܂łBx[XC R[h쐬ɏI܂B");
		return 0;
	}
}
else{
	$stream_used = $opts{ucmstream};
	$logger->log_print_message(" 󂯓n UCM Xg[gpĂ܂: $stream_used");
}

$logger->log_print_message(" pvob  UCM Xg[܂: $stream_used");
$logger->log_print_message(" vWFNg ID  :$opts{projectid} ł");

my @members;
if ( $opts{add_composed_of_baseline} ) {
	#Convert these pairs into {bl=>"", pvob=>""} pairs;
	$logger->log_print_message("Found -add_composed_of_baseline IvV");
	for ( my $i = 0; $i < @{$opts{add_composed_of_baseline}}; $i++ ) {
		my $bl = $opts{add_composed_of_baseline}[$i];
		my $pvob = $opts{add_pvob_or_location}[$i];
		$logger->log_print_message("x[XC:$bl  pvob_or_loc:$pvob ");
		push @members, {bl=>$bl,bl_pvob=>$pvob};
	}
}

# Check if -nodiffbl or -activity  specified
if (@{$opts{activity}} or ! $opts{diffbl}){
	#activities were passed to the script on the commandline or in a file.
	#no need to diffbl in this case.
	$logger->log_print_message("-activityA-from ܂ -nodiffbl w肳Ă܂: diffbl ̓XLbv܂");
	my $result = create_new_record(session => $session, stream => $stream_used, baseline => $opts{baseline}, projectid => $opts{projectid}, pvob => $opts{pvob}, user => $opts{user}, diffbl_list => \@{$opts{activity}}, members => \@members, ucm=>0 );
	if (! $result){
		$logger->log_print_error(__LINE__ . " x[XC R[h쐬ł܂łB");
		return 0;
	}
}
else {

	# If composite baseline
	# get add_composed_of_baseline baselines, if no record exists for them, then create them.
	# When calling create_baseline_records.pl decrement the -create_composed_of_baselines_level by 1.
	# If this level is 0 create_baseline_records.pl should fail for composite baselines.
	# diffbl to get only local activities to add to "this" baseline record.
	#
	#### Build a query to find the baseline record to diffbl against
	# TODO add -add_composed_of_baseline option parsing for NON-UCM usage, these need to be in the form ALM0001234

	my $blcc = new get_baseline_details(baseline=>$opts{baseline},pvob=>$slashvob);
	if (! defined $blcc ) {
		$logger->log_print_error(__LINE__ . "x[XC̏ڍׂ擾ł܂ baseline=>$opts{baseline},pvob=>$slashvob: $@"); 
		return 0;
	}
	# Save the component, if it's not passed in on the commandline
	if ( ! defined $opts{component} ) {
		# ($opts{component}) = $blcc->get("component");
		($opts{component}) = ($blcc->get("_R|[lg"))->[0];
		$opts{component} =~ s/@.*//;
	}

	# Test to see if the baseline stream matches the passed in UCM stream.
	# my($stream_from_baseline) = $blcc->get("stream");
	my($stream_from_baseline) = ($blcc->get('_Xg['))->[0];
	$stream_from_baseline =~ s/@.*//;
	if ( defined $opts{stream} and $opts{stream} ne $stream_from_baseline ) {
		$logger->log_print_error(__LINE__ . qq{: -stream $opts{stream} does not match stream from $opts{baseline} of $stream_from_baseline});
		$logger->log_print_error(__LINE__ . qq{: Substituting $stream_from_baseline});
	}
	$opts{stream} = $stream_from_baseline;

	#Get a list of baselines from the component to test for a match from CQ.
	my $cmd = qq{cleartool lsbl -s -component $opts{component}\@$slashvob -stream $opts{stream}\@$slashvob};
	my ($retval, $output) = $ut->ut_shell($cmd);
	my %comp_baselines = map {$_ => 1 } split /\n/,$output;
	unless ($retval ) {
		$logger->log_print_error(__LINE__ . " $cmd ̎s̃G[: $retval");
		return 0;
	}

	my $is_composite = 0;
	# if there are add_composed_of_baseline baselines this is a composite.
	my @loc_members = $blcc->get("members");
	@members = (@members, @loc_members);
	for my $member (@members) {
		$is_composite++;
		my $member_bl = $member->{bl};
		my $member_pvob = $member->{bl_pvob};
		$logger->log_print_message(__LINE__ .  " o[̃x[XC̒: bl = $member_bl  pvob = $member_pvob");
		my ($slashvob, $noslashvob ) = &slash_noslash($member_pvob);

		# Just a quick check for the record - without all the details. top=>1
		my $blcq = new get_baseline_record_details(session=>$session, baseline=>$member_bl, pvob=>$noslashvob,top=>1);

		if ( ! defined $blcq ) {
			#member baseline record doesn't exist;
			if ( $opts{create_composed_of_baselines} ) { # Yes create member baselines
				if ( $opts{create_composed_of_baselines_level} > 0 ) {
					my $new_level = $opts{create_composed_of_baselines_level} - 1;
					#  -create_composed_of_baselines_level $new_level
					# TODO investigate projectid
					# TODO investigate logfile naming
					my $cmd = qq|$perl "$0" |;
					for my $opt (@required_opts) {
						next if $opt =~ /baseline/;
						$cmd .=  qq| -$opt "$opts{$opt}" |;
					}

					$cmd .=
					qq|-create_composed_of_baselines_level $new_level -create_composed_of_baselines | .
					qq|-baseline "$member_bl" | . 
					qq|-pvob "$member_pvob" | .
					qq|-logfile "$opts{logfile}.$member_bl" | .
					qq| |;
					$logger->log_print_message(__LINE__ . ": $cmd");
					my ($retval, $output) = $ut->ut_shell($cmd);
					unless ($retval ) {
						$logger->log_print_error(__LINE__ . " create_baseline_record ̎s̃G[: $retval");
						$logger->log_print_error(__LINE__ . " $retval");
						return 0;
					}
				}
				else {
					# 
					$logger->log_print_error(__LINE__ . "$member_bl $member_pvob : create_composed_of_baselines_level ̓[ɒB܂Ao[̃x[XC R[h쐬ł܂"); 
					return 0;
				}
			}
			else {
				$logger->log_print_error(__LINE__ . "$member_bl $member_pvob : x[XC R[h݂܂B-create_composed_of_baselines w肳Ă܂Bo[̃x[XC R[h쐬ł܂"); 
			}
		}
		else {
			#my ($member_bl) = $blcq->get("baseline");
			#my ($member_pvob) = $blcq->get("PVOB_OrLocation");
		}
	}

	my $queryDef = $session->BuildQuery("ALMBaseline");
	my $failure = $queryDef->BuildField("Name");
	$failure = $queryDef->BuildField("UCM_Stream");
	$failure = $queryDef->BuildField("history.action_timestamp"); # changed 03/13/03
	$failure = $queryDef->BuildField("PromotionLevel"); #changed 03/14/03
	$failure = $queryDef->BuildField("PVOB_OrLocation");
	$failure = $queryDef->BuildField("Obsolete");                              
	  
	my $operator = $queryDef->CreateTopNode($CQPerlExt::CQ_BOOL_OP_AND); 
	$operator->BuildFilter("PVOB_OrLocation", $CQPerlExt::CQ_COMP_OP_EQ, ["$noslashvob"]);
	$operator->BuildFilter("UCM_Stream", $CQPerlExt::CQ_COMP_OP_EQ, ["$stream_used"]);
	$operator->BuildFilter("history.action_name", $CQPerlExt::CQ_COMP_OP_EQ, ["Submit"]);
	$operator->BuildFilter("PromotionLevel", $CQPerlExt::CQ_COMP_OP_NEQ, ["REJECTED"]); # changed 03/14/03 
	$operator->BuildFilter("Obsolete", $CQPerlExt::CQ_COMP_OP_NEQ, ["Yes"]);

	my $queryfielddefs = $queryDef->GetQueryFieldDefs(); 
	my $baselinefield = $queryfielddefs->ItemByName("history.action_timestamp");# changed 03/13/03
	$baselinefield->SetSortType($CQPerlExt::CQ_SORT_DESC); 
	$baselinefield->SetSortOrder(1); 

	eval {$resultSet = $session->BuildResultSet($queryDef)};
	# Write to log file
	if ($@) {
	$logger->log_print_error(__LINE__ . "NG[̃rhɎs܂: $@"); 
	return 0;
	}
	else {$logger->log_print_message(" f[^x[XNG[ʂ𐶐Ă܂");}


	my $num_records = $resultSet->ExecuteAndCountRecords(); 
	# Find the first record within ClearQuest
	for (my $i = 0; $i < $num_records; $i++){ 
		$resultSet->MoveNext;
	    if ($resultSet->GetColumnValue(2) eq "$opts{baseline}") {
			$logger->log_print_error(__LINE__ . "$opts{baseline} ̃x[XC R[h͊ɑ݂܂B");
			$logger->log_print_error(__LINE__ . "dx[XC R[h쐬邱Ƃ͂ł܂BZbVI܂B");
			CQPerlExt::CQSession_Unbuild($session);
			return 0;
	    } else {	
		# Only concider baselines that are from the component
		my $current_bl = $resultSet->GetColumnValue(1);
		if ( defined $comp_baselines{$current_bl}) {
		$oldbaseline = $current_bl;
		$oldPVOB = $resultSet->GetColumnValue(5);                                  #thl2
		$logger->log_print_message("O̗ǂx[XC $oldbaseline łB");
		$i = $num_records;
		}
	     } 
	}

	$logger->log_print_message("  PVOB l $oldPVOB ł");

	 
	## Determine if previous Baseline records were found in ClearQuest database

	if ($num_records eq "0"){
		$logger->log_print_message(" Õx[XC R[h܂B");
		$logger->log_print_message(" Õx[XC R[h UCM ̎gp\ VOB ȂŁAx[XC R[hrhĂ܂B");

		## Find the project associated with the foundation baseline
		$project = get_project( stream => $stream_used, pvob => $slashvob);
		if (! $project){
			$logger->log_print_error(__LINE__ . "pvob ̃Xg[ vWFNg𔻕ʂł܂łBx[XC R[h쐬ɏI܂B");
			return 0;
		}
		
		##Find the modifiable components of the new baseline record
		$mod_comps= get_modcomponents(project => $project, pvob => $slashvob);
		if (! $mod_comps){
			$logger->log_print_error(__LINE__ . "vWFNg̕ύX\ȃR|[lg𔻕ʂł܂łBx[XC R[h쐬ɏI܂B");
			return 0;
		}
		
		## Get "Seed" of Baseline record
		$oldbaseline = get_seed( stream => $stream_used, pvob => $slashvob);
		if (! $oldbaseline){
			$logger->log_print_error(__LINE__ . "gpĂV[h x[XC𔻕ʂł܂łBx[XC R[h쐬ɏI܂B");
			return 0;
		}
		
		## Iterate through all the foundation baselines to find the ones with modifiable components
		$foundbaseline = found_baselines(oldbaseline => $oldbaseline);                            #thl2 
		if (! $foundbaseline){
			$logger->log_print_error(__LINE__ . "V[h x[XC𔻕ʂł܂łB");
			$logger->log_print_error(__LINE__ . "x[XC R[h쐬ɏI܂B");
			return 0;
		}
		## Get the label status of the new baseline
		$baseline_status = get_status( pvob => $slashvob, baseline => $opts{baseline} );
		if (! $baseline_status){
			$logger->log_print_error(__LINE__ . "x[XC̃Xe[^X𔻕ʂł܂łBx[XC R[h쐬ɏI܂B");
			return 0;
		}
		my $diffbl_list ;                                                                         #thl1
		if ($baseline_status =~ /S/i || $baseline_status =~ /CN^/i){			
			# Iterate over each foundation baseline to find the total activities
			$diffbl_list = total_activities(foundbaseline=>$foundbaseline, %opts);            #thl2 
			}
		if (! $diffbl_list and !$is_composite) {                                                                     #thl2
		   $logger->log_print_warning(__LINE__ . "ANeBreB𔻕ʂł܂ł:") ;       #thl2
		   $logger->log_print_warning(__LINE__ . "x[XC: $opts{baseline}");                 #thl2
		   $logger->log_print_warning(__LINE__ . "x[XC R[h쐬ɏI܂B"); #thl2
		   return 0 ;                                                                             #thl2
		}                                                                                         #thl2
		$logger->log_print_message(" fo[ꂽׂẴANeBreBA̓x[XC $opts{baseline} ̃R|WbgłB"); #thl2
		# Get rid of duplicates 
		my %diffbl_list;
		map { $diffbl_list{$_}++ } @$diffbl_list;
		my @diffbl_list = sort keys %diffbl_list;
		$logger->log_print_message(" ANeBreBύX܂: @diffbl_list");
		my $result = create_new_record(session => $session, stream => $stream_used, baseline => $opts{baseline}, projectid => $opts{projectid}, pvob => $opts{pvob}, user => $opts{user}, diffbl_list => \@diffbl_list, members=>\@members);
		if (! $result){
			$logger->log_print_error(__LINE__ . "x[XC R[h쐬ł܂łB");
			return 0;
		}
	} # end of no previous baseline found in CQ 


	else { 
		## Previous baseline found in ClearQuest database
		$logger->log_print_message(" Õx[XC R[h UCM ̎gp\ VOB gpāAx[XC R[hrhĂ܂B");
		
		## Using the diffbl
	$oldPVOB="\\".$oldPVOB if $Config{osname} =~ /Win32/; # Windows only issue
		# If this is a composite then add nrecurse=>1
		my $diffbl_list = get_activities( BL1 => $oldbaseline, BL2 => $opts{baseline}, pvob1 => $slashvob, pvob2 => $oldPVOB, nrecurse=>$is_composite);
		if (! defined $diffbl_list and ! defined $is_composite){
			$logger->log_print_warning(__LINE__ . "ANeBreB𔻕ʂł܂ł:");
			$logger->log_print_warning(__LINE__ . "x[XC: $oldbaseline");
			$logger->log_print_warning(__LINE__ . "x[XC: $opts{baseline}");
			$logger->log_print_warning(__LINE__ . "x[XC R[h쐬ɏI܂B");
			$logger->log_print_warning(__LINE__ . " ꂪ߂Ẵx[XC R[h̏ꍇA͂̃vWFNǵuV[hvx[XCłǍx͖܂B");
			return 0;
		}

# Determine the number of activities 
#my $totalactvities = $@diffbl_list+1; 

		$logger->log_print_message(" x[XC $oldbaseline ƃx[XC $opts{baseline} ԂŃfo[ꂽAׂẴANeBreB܂B");	
		for my $key (keys %${diffbl_list}){
			$logger->log_print_message(qq{"$key" => "$diffbl_list->{$key}"});
		}


		$logger->log_print_message(" x[Xуfo[ R[h̃Rgr[^ɂȂ邷ׂẴANeBreB܂");
		$logger->log_print_message(" : ̃ANeBreBɂ̓Rgr[^ɂȂANeBreBȂAU̎s lsact -contrib ŕ񍐂ꍇ܂A͖܂");
		my $all_diffbl_list = find_contribs(pvob => $slashvob, ACT => $diffbl_list);
		$logger->log_print_message(" x[Xуfo[ R[h̃Rgr[^ɂȂ邷ׂẴANeBreB܂");
		
		# Get rid of duplicates 
		my %diffbl_list;
		map { $diffbl_list{$_}++ } @$all_diffbl_list;
		my @diffbl_list = sort keys %diffbl_list;






		$logger->log_print_message(" ANeBreBύX܂: @diffbl_list");
		
		#execute the baseline record creation
		my $result = create_new_record(session => $session, stream => $stream_used, baseline => $opts{baseline}, projectid => $opts{projectid}, pvob => $opts{pvob}, user => $opts{user}, diffbl_list => \@diffbl_list, members => \@members);
			if (! $result){
				$logger->log_print_error(__LINE__ . " x[XC R[h쐬ł܂łB");
				return 0;
		}
		
			
	} # end of $num_records loop
} # End if 

$logger->log_print_success(" Create_baseline_record XNvg܂Bx[XC R[h쐬܂B");
return 1;

} #end of sub main



####Subroutines for Program#####
# Subroutine to display the usage

sub Usage {
my $message ="
gp@: $progname -user <username> -pw <password> -dbname <database name> -dbset <database set> -pvob <pvob> -baseline <baseline> -projectid <project id> [-logfile <logfile>] [-ucmstream <ucmstream>] [-create_composed_of_baselines] [-test] [-help] [-activity <activity>] [-from_file <file of activities>]\n".
"\n".
"$progname $VERSION :CQ API gpăx[XC R[h쐬邽߂̃[eBeB\n".
"\n".
" pvob  ucmstream gpAALM ̈ȑÕx[XC R[h diffbl s܂B\n".
" diffbl 猟oANeBreB́AVKx[XC R[hɒǉ܂B\n".
" UCM R|[lggpꍇɂ́AK؂ȃx[XC̈ˑ֌WR|Wbg x[XCgpKv܂Bx[XCɃpXꂽ̂R|Wbg x[XCł邩ǂ́AXNvgIɔʂ܂B\n".
"\n".
"K{:\n".
"  -user <user>       ClearQuest [U[̃OC IDB\n".
"  -pw <password>     ClearQuest [U[̃pX[hB\n".
"  -dbname <dbname>   f[^x[Xւ̃XL[} |WgڑB\n".
"  -dbset <dbset>     [U[ f[^x[XڑB\n".
"  -pvob <\\pvob>     vWFNg VOB ̖OB\n".
"  -baseline <blname>  ClearCase x[XC̖OB\n".
"  -projectid <id>    ALM ̃vWFNg IDB\n".
"\n".
"IvV:\n".
"  -create_composed_of_baselines	R|WbgXNvgɃpXƁAo[ x[XC \"x[XC\\" tB[hɒǉ܂B̃IvVw肷ƁAo[ x[XC R[h܂݂ȂꍇA쐬邱Ƃł܂B́AċAI $progname ĂяoƂɂĐs܂BŎw肳ꂽ̂ƓuK{vIvV́A$progname ̍ċAĂяoɃpX܂B\n".
"    : ̃R|Wbg x[XC܂ރR|Wbg x[XĆA݂̓T|[gĂ܂Bʏ̃x[XĈ݂܂ނ悤ɃR|Wbg𒲐ĂB\n".
"\n".
"  -logfile <logfile> R\[ɕ\ƋɁAw肳ꂽt@Cɏo͂OƂċL^܂B\n".
"  -test              p[^Ŋ{xɃpXꂽ̂eXg邾ŏI܂BR[h͂܂쐬܂B\n".
"  -help              ̃bZ[Wo͂ďI܂B\n".
"\n".
" UCM IvV:\n".
"ȉ̃IvVgpāA UCM  UCM V~[g܂B\n".
"  -nodiffbl     UCM x[XCłȂ߁Adiffbl (diff x[XC) s܂B\n".
"               L -activity ܂ -from_file IvVtăANeBreBw肵܂B\n".
"  -ucmstream <stream>   V~[g ucmstream ̖OŃpXKv܂B\n".
"  -activity <activity> [-activity <activity>]   쐬ꂽx[XC R[hɁAw肳ꂽANeBreBǉ܂B\n".
"  -from_file <file>     diffbl ̑ɁAUCM ƈꏏɎgp܂B\n".
"                        sƂ 1 Xgꂽ ALM ANeBreBƋɁAt@Cw肳A\n".
"                        ̃ANeBreBA쐬ꂽx[XC R[hɒǉ܂B\n".
"                        stH[}bg:\n".
"                        ALM00000003\n".
"                        ALM00000777\n".
"                        ALM00000453\n".
"\n".
" UCM V~[g R|Wbg x[XC\n".
"      R|Wbg x[XĆA\"x[XC\\" tB[hɒlw肷邱ƂɂăV~[gł܂B\n".
"  -add_composed_of_baseline <nonUCM-baseline> \n".
"  -acc_pvob_or_location <equiv_to_pvob>\n".
"         -add_comp  -add_pvob ͏ȗł܂B\n".
"         \"x[XC\\" tB[hɒǉx[XCƂɁÃyAɂȂIvVǉ܂B\n".
"         : -add_comp existing_nonUCM_baseline_record_01 -add_pvob nonUCM_equive_to_pvob\n".
"         x[XCƂɗ̃IvVKvłB\n".
"\n".
": -activity  -from_file ̏ꍇApXANeBreB܂ ALM ɑ݂Kv܂BANeBreB̃Xe[^X܂͏Ԃ̓`FbN܂B\n".
"\n".
"\n";

print $message;
	
return 0;
}



# Get the stream based on pvob and baseline
#
sub get_stream {
    	my %opts= @_;
        my $cmd = qq(cleartool lsbl "$opts{baseline}\@$slashvob"); 
	my ($retval, $output) = $ut->ut_shell($cmd);
   
   	unless ($retval){
		$logger->log_print_error(__LINE__ . " Xg[`Ă܂BXg[܂ PVOB mFĂB");
		$logger->log_print_error(__LINE__ . " Cleartool R}h $cmd ̓Xg[̎擾Ɏs܂B");
		$logger->log_print_error(__LINE__ . " Cleartool o: $output");
		$logger->log_print_error(__LINE__ . "Xg[Ȃł́Ax[XC쐬ł܂");
		return 0;
		my %opts = @_;
    	}
	# $output =~ /stream: (.*?)@./ms;
	my $match = quotemeta('Xg[');
        $output =~ /$match: (.*?)@./ms;
	my $stream = $1;
	chomp $stream;
	return $stream;
}

# Find Project for the Baseline Record  
#
sub get_project {
	my %opts = @_;
#	my $cmd = qq(cleartool lsstream -fmt "%[project]p" "$opts{stream}\@$slashvob");
        my $cmd = qq(cleartool lsstream -fmt "%[project]p" "$stream_used\@$slashvob");
	my ($retval, $output) = $ut->ut_shell($cmd);
        my $project;
       
        unless ($retval){
		$logger->log_print_error(__LINE__ . " Cleartool R}h $cmd ̓vWFNg̎擾Ɏs܂B");
		$logger->log_print_error(__LINE__ . "vWFNgȂł́Ax[XC쐬ł܂");
		return 0;
        }
        while (defined ($project = $output)){
 		chomp $project;
		$logger->log_print_success(" vWFNg: $project");				
		return $project; 
    	}   	  	       	 	
} 

# Given two baselines diff them and get all the activities.
#
sub get_activities {

	my %opts = @_;
	my $nrecurse;
	$nrecurse = "-nrecurse" if $opts{nrecurse};
        my $cmd   =
      	 qq(cleartool diffbl $nrecurse -activities "$opts{BL1}\@$opts{pvob1}" "$opts{BL2}\@$opts{pvob2}"); #thl2
	my ($retval, $output) = $ut->ut_shell($cmd);
	unless ($retval) {
		$logger->log_print_warning(__LINE__ . " Cleartool R}h $cmd ̓ANeBreB̎擾Ɏs܂B");
		$logger->log_print_warning(__LINE__ . "ANeBreBȂł́Ax[XC쐬ł܂");
		$logger->log_print_warning(__LINE__ . " ꂪ߂Ẵx[XC R[h̏ꍇA͂̃vWFNǵuV[hvx[XCłǍx͖܂");
		return 0;
	}
	my %lact;
	my @list = split /\n/, $output;
	foreach my $act (@list) {
		chomp $act;
		if ( $act =~ />>\s([^\s]*)\s"(.*?)"/ ) {
			## SP
			my $a = $1;            # The RATLC Activity id
			my $b = $2;            # The Activity Headline
			$a =~ s/\@.*//;
                        $lact{$a} = $b;
		}
	}
	return \%lact;
}
# Given a set of activities, find contributing activities.
#
sub find_contribs {
    my %opts = @_;
    my @return;
    for my $act ( keys %{ $opts{ACT} } ) {
	next if $seen{"$act\@$slashvob"}++;
	my $match1 = quotemeta('x[X');
	my $match2 = quotemeta('fo[');
	next if $opts{ACT}->{$act} =~ /^$match1/i;
	next if $opts{ACT}->{$act} =~ /^$match2/i;
	
### SP
	my $cmd;
	if ($act !~ /\@/){
        	$cmd     = "cleartool lsact -contrib $act\@$slashvob";
	}
	else {
		$cmd     = "cleartool lsact -contrib $act";
	}
	my ($retval, $output) = $ut->ut_shell($cmd); 
        chomp $output; 
        $output =~ s/"//g;    # results enclosed in quotes "..." (?)
        # if ( $output =~ /is not a deliver or rebase integration activity/ ) {
        my $match = quotemeta('̓fo[܂̓x[X CeO[V ANeBreB');
        if ( $output =~ /$match/ ) {
            push @return, $act;
        }
        elsif ( $output =~ /""/ ) {    # Some activities are empty?
            push @return, $act;
        }
        else {
		#Recursion
            	my @list = split /\n/,$output;
		my %list;
		for my $act ( @list ) {
			chomp $act;
			$act =~ s/"//g;
			$list{$act} = "";
	        }
            	push @return, @{ find_contribs( pvob => $slashvob, ACT => \%list) };
	}
    }
    return \@return;
}
# Find Label Status of Baseline Record  
#
sub get_status {
	my %opts = @_;
	my $cmd = qq(cleartool lsbl -fmt "%[label_status]p" "$opts{baseline}\@$slashvob");
	my ($retval, $output) = $ut->ut_shell($cmd);
        my $status;
        unless ($retval){	
		$logger->log_print_error(__LINE__ . " Cleartool R}h $cmd ̓x̏ԏ̎擾Ɏs܂B");
		$logger->log_print_error(__LINE__ . "x̏ԏȂł́Ax[XC쐬ł܂");
		return 0;
        }
        while (defined ($status = $output)){
 		chomp $status;
		$logger->log_print_success(" x̏: $status");
		return $status;
    	}     	  	       	 	
}  
# Find Foundation (Seed) Baseline(s) Record 
#
sub get_seed {
        my %opts = @_;
        # thl2 we have to use -long because -fmt "%[found_bls]p" does not return the baseline selector, just its the name. 
#	my $cmd = qq(cleartool lsstream -long "$opts{stream}\@$slashvob");  
        my $cmd = qq(cleartool lsstream -long "$stream_used\@$slashvob"); 
	my ($retval, $output) = $ut->ut_shell($cmd);
        my $seed;
        my $infbls = 0 ;  #thl2
	my @fseed;
        unless ($retval){
		$logger->log_print_error(__LINE__ . " Cleartool R}h $cmd ̓V[h R[h`̎擾Ɏs܂B");
		$logger->log_print_error(__LINE__ . "V[h R[hȂł́Ax[XC쐬ł܂");
		return 0;
        }
        # split on the new line chars
	my @output = split(/\n/,$output) ;          #thl2
        # 
        loop: for (my $i=0;$i<=$#output;$i++) {     #thl2                                     
           $seed = $output[$i] ;                    #thl2
           chomp $seed ; 
           $seed = join(" ",split(" ",$seed)) ;     #thl2
           my $match1 = quotemeta('{x[XC');
           if ($seed =~ /^$match1:/) { #thl2 Found the start of the foundation baseline section
              $infbls = 1 ;                         #thl2
              next loop ;                           #thl2
           }                                        #thl2
           elsif (($infbls) && ($seed =~ /:$/)) {   #thl2 This marks the end of the foundation baseline section
              last loop ;                           #thl2
           }                                        #thl2
           elsif ($infbls) { #thl2 still in the foundation baseline sections. Grab the bl
              $logger->log_print_success(" {x[XC: $seed"); 
              push @fseed, $seed;                                        
           }
        } # end loop 
    	return \@fseed;  	       	 	
}  
# Find Components for the Baseline Record  
#
sub get_components {
	my %opts = @_;
	my $cmd = qq(cleartool lsbl -fmt "%[component]p" "$opts{baseline}\@$slashvob");
	my ($retval, $output) = $ut->ut_shell($cmd);
        unless ($retval){
		$logger->log_print_error(__LINE__ . " Cleartool R}h $cmd ̓R|[lg̎擾Ɏs܂B");
		$logger->log_print_error(__LINE__ . "R|[lgȂł́Ax[XC쐬ł܂");
		return 0;
        }
	my @comp_list;
	my @list = split /\n/, $output;
	foreach my $comp (@list) {
		chomp $comp;
		push @comp_list, $comp;
	}
    	# Write to log file
	$logger->log_print_success("w肳ꂽx[XC̃R|[lg: @comp_list");
	return \@comp_list;  	  	       	 	
} 
# Find modifiable components for the Baseline Record  
#
sub get_modcomponents {
	my %opts = @_;
	my $cmd = qq(cleartool lsproject -fmt "%[mod_comps]p" "$opts{project}\@$slashvob");
	my ($retval, $output) = $ut->ut_shell($cmd);
        unless ($retval){
		$logger->log_print_error(__LINE__ . " Cleartool R}h $cmd ͕ύX\ȃR|[lg̎擾Ɏs܂B");
		$logger->log_print_error(__LINE__ . "ύX\ȃR|[lgȂł́Ax[XC쐬ł܂");
		return 0;
        }
	my @modcomp_list;
	my @list = split /\n/, $output;
	foreach my $modcomp (@list) {
		chomp $modcomp;
		push @modcomp_list, $modcomp;
	}
	$logger->log_print_success("w肳ꂽx[XC̕ύX\ȃR|[lg: @modcomp_list");
	return \@modcomp_list;  	  	       	 	
} 

# Find Promotion Level for the Baseline Record  
#
sub get_promotionlevel { 
	my %opts = @_;
	my $cmd = qq(cleartool lsbl -l "$opts{baseline}\@$slashvob"); #021703
	my ($retval, $output) = $ut->ut_shell($cmd);
        unless (defined $retval){
		$logger->log_print_error(__LINE__ . " Cleartool R}h: $cmd"); 
		$logger->log_print_error(__LINE__ . "v[V x̎擾Ɏs܂");
		$logger->log_print_error(__LINE__ . "v[V xȂł́Ax[XC쐬ł܂");
		return 0;
        }
    my $match = quotemeta('v[V x');
    $output =~ /$match: (.*?)$/ms;
	my $plevel = $1;
	chomp $plevel;
	$logger->log_print_success("v[V x $plevel ł");
	return $plevel;   	 	
} 
# Iterate through all the foundation baselines to find the ones with modifiable components

sub found_baselines{
	my %opts = @_;
	my @oldbaseline = $opts{oldbaseline};
	foreach my $item(@$oldbaseline){
                ##The format of $item is expected to be a blank delimited list with these elements   thl2
                 # baseline_selector component_selector [(modifiable)]                               thl2
                 # parse out the baseline tag and it's pvob                                          thl2
                my ($bs_selector,undef,undef) = split(/ /,$item) ;                                   #thl2
                my ($foundbaseline,$pvob) = split(/@/,$bs_selector) ;                                #thl2 
                ##Find the components of the foundation baseline record                              
                my $comps= get_components(baseline => $foundbaseline, pvob => $pvob);                #thl2
		if (! $comps){
			$logger->log_print_error(__LINE__ . " x[XC $item ̃R|[lg𔻕ʂł܂łB");
			return 0;
		}
		
		## Determine if the component is a modifiable one		
		foreach my $e(@$mod_comps) {$union{$e}= 1}
		foreach my $e(@$comps){
		  if($union{$e}) {$isect{$e}=1}
			$union{$e} =1;
		} # end of foreach loop

		@union= keys %union;
		@isect =keys %isect;
		# Keep track of the foundation baseline that have modifiable components
		if (@isect ne ""){
		push  @foundbaseline, $bs_selector;                                                   #thl2
		} # end of tracking baselines
	} # end of iterating through all the foundation baselines
        return \@foundbaseline;
}
# Iterate over each foundation baseline to find the total activities
sub total_activities{
	my %opts = @_;
	foreach my $item(@{$opts{foundbaseline}}){	
	    if ($item ne $opts{baseline} ){
                # split out the basline name and pvob and send them to 
                # get_activities as separate opts
                my ($bn,$pv) = split(/@/,$item) ; #thl2 
		## Using the diffbl
		my $diffbl_list = get_activities( BL1 => $bn, BL2 => $opts{baseline}, pvob1 => $pv, pvob2 => $slashvob); #thl2
		if (! $diffbl_list){
			$logger->log_print_error(__LINE__ . " x[XC̃ANeBreB𔻕ʂł܂łB");
			return 0;
		}
		$logger->log_print_message(" x[XC $itemAx[XC $opts{baseline} ̃ANeBreB̍v܂");
		$result = my $all_diffbl_list = find_contribs( ACT => $diffbl_list, pvob => $slashvob);
		if (! $result){
			$logger->log_print_error(__LINE__ . " Rgr[^ɂȂANeBreB𔻕ʂł܂łB"); 
			return 0;
		}
		$logger->log_print_message(" Rgr[^ɂȂANeBreB̍v܂");
		## Get rid of duplicates 
		my %diffbl_list;
		map { $diffbl_list{$_}++ } @$all_diffbl_list;
		my @diffbl_list = sort keys %diffbl_list;
		push (@list, @diffbl_list);
	    } # end of checking to see if oldbaseline equals new baseline
	} # end of the foreach loop

	# Final check for duplicates on all foundation baselines found
	%seen = ();
	foreach my $item (@list) {
    		$seen{$item}++;
	}
	my @uniq = keys %seen;
        return \@uniq ;
}

sub create_new_record {
	my %opts = @_;
	$opts{ucm} = 1 if (!defined $opts{ucm});
	my @diffbl_list = @{$opts{diffbl_list}};
	my ($status,$item); #thl3
	my $session = $opts{session};
	my @members = @{$opts{members}};
	# TODO if there are @members then put these on the record.
	
	## Find the promotion level of the baseline
	if ($opts{ucm}){
		$result = $promotion_level = get_promotionlevel(baseline => $opts{baseline}, pvob => $slashvob);
		if (! $result){
				$logger->log_print_error(__LINE__ . " x[XC $opts{baseline} ̃v[V x𔻕ʂł܂łBx[XC R[h쐬ɏI܂B");
				return 0;
			}
			
		if ($promotion_level eq "REJECTED"){
			$logger->log_print_error(__LINE__ . " x[XC $opts{baseline} p܂B");
			$logger->log_print_error(__LINE__ . " pꂽ x[XC̃x[XC R[h쐬ł܂");
			return 0;
		}
	}
	## Build the Baseline record
	my $entity = $session->BuildEntity("ALMBaseline");

$logger->log_print_message("Xg[̒l $stream_used łB");
$logger->log_print_message(" vWFNg ID ̒l $opts{projectid} łB");
	
	#### Set field values for pvob and baseline
	my $failure = $entity->SetFieldValue("Project", "$opts{projectid}");
        $failure = $entity->SetFieldValue("PVOB_OrLocation", $noslashvob);
        $failure = $entity->SetFieldValue("Name", "$opts{baseline}");
        $failure = $entity->SetFieldValue("PromotionLevel", "$promotion_level");
        ### $failure = $entity->SetFieldValue("Owner", "$userLogin");
        $failure = $entity->SetFieldValue("UCM_Stream", "$stream_used"); 

$logger->log_print_message(" tB[hl̐ݒ󂯓n܂");
 
        # Diffbl_list now contains the complete list of activities/contributing activities in the baseline.
	if (@diffbl_list) {
		foreach $item (@diffbl_list) {
	  		$entity->AddFieldValue("Activities",$item);
			$logger->log_print_message("ANeBreBǉ܂: $item");
		} # end of adding new CRs
	} # end of diffbl

	if (@members) {
		$logger->log_print_message(" o[ Xg IF 󂯓n܂");
		for my $item (@members) {
			my $bl = $item->{bl};
			my $pvob = $item->{bl_pvob};
			my ($slash, $noslash) = slash_noslash($pvob);
			$logger->log_print_message(" o[̒l '$noslash $bl' ł");
			$entity->AddFieldValue("ComposedOfBaselines", "$noslash $bl");
		} 
	} # end of members
								
	## New Baseline ID
	$logger->log_print_message(" VKx[XC R[h $opts{baseline} ؂Ă܂");
	eval {$status = $entity->Validate()}; #thl3
	if (($@) || ($status)) {              #thl3
		$logger->log_print_error(__LINE__ . " VKx[XC R[h͌؂Ɏs܂: $@");
                $logger->log_print_error(__LINE__ . "߂Xe[^X $status ł"); #thl3
    		$entity->Revert();
   		CQPerlExt::CQSession_Unbuild($session);
    		return 0;
	}
        else {
           $logger->log_print_message(__LINE__ . " VKx[XC R[h؂܂: $@");
        }
	$status = "" ; #thl3
	$logger->log_print_message(" VKx[XC R[h $opts{baseline} R~bgĂ܂");
	eval {$status = $entity->Commit()}; #thl3
	if (($@) || ($status)) {            #thl3
		$logger->log_print_error(__LINE__ . " VKx[XC R[h̓R~bgɎs܂: $@");
                $logger->log_print_error(__LINE__ . "߂Xe[^X $status ł"); #thl3
    		$entity->Revert();
		CQPerlExt::CQSession_Unbuild($session);
		return 0;
	}
	$entity = undef;
	$logger->log_print_message(" VKx[XC R[h $opts{baseline} 쐬܂");	
	CQPerlExt::CQSession_Unbuild($session);
	return 1;		
} 

#
# get_member_baselines
# Passed a baseline and pvob, return a list of members.  
#
sub get_member_baselines {
	my %lopts = @_; 
	my $cmd = qq{cleartool lsbl -fmt "\%[depends_on]Cp" $lopts{baseline}\@$lopts{pvob}};
	my $results = `$cmd`;
	my @members;
	if ($results ) {
		@members = split ",\s", $results;
	}
	@members;
}

#
# get_baseline_details
# Passed in a baseline and pvob, return via get method values
#

package get_baseline_details;
# Given a baseline, this package returns details on it via get("key");
sub new { 
	my $class = shift;
	my $self = {@_};
	bless($self, $class);
	if ( defined $self->{baseline} ) {
		$self->get_baseline_details();
	}
	return $self;
}

sub get_baseline_details {
	my $self = shift;
	my $cmd = qq{cleartool lsbl -l $self->{baseline}\@$self->{pvob}};
	my @lines = `$cmd`;
	# TODO test results of $cmd
	chomp @lines;
	for my $line (@lines) { #get rid of leading and trailing space in lines
		$line =~ s/^\s*//;
		$line =~ s/\s*$//;
	}
	#loop through the lines and parse out the data
	OUTER: while (my $line = shift @lines ) {
		next OUTER if $line !~ /:/;
		my ($what,$data) = split /:\s*/, $line, 2;
		my $match1 = quotemeta('ύXZbg');
		if ($what =~ /$match1/ ) {
			while (my $l = shift @lines) {
				if ($l =~ /:/) { unshift @lines,$l; next OUTER; }
				push @{$self->{"activities"}},$l;
			}
		}
		my $match2 = quotemeta('ˑ֌W');
		if ($what =~ /$match2/ ) {
			while (my $l = shift @lines) {
				if ($l =~ /:/) { unshift @lines,$l; next OUTER; }
				#gui_11_26_2007_01@\qvmw388_pvob_01 (qvmw388_gui@\qvmw388_pvob_01)
				my ($bl,$bl_pvob,$comp,$comp_pvob) = ($l =~ /^(.*)\@(.*)\s+\((.*)\@(.*)\)/ );
				my $members = { 'bl' => $bl, 'bl_pvob' => $bl_pvob, 'comp' => $comp, 'comp_pvob' => $comp_pvob };
				push @{$self->{"_members"}},$members;
			}
		}
		else {
			push @{$self->{"_$what"}} , $data;
		}
	}
	return $self;
}

sub get {
	my $self = shift;
	my $what = shift;
	return $self->{$what} if ( exists $self->{$what} ) ;
	return @{$self->{"_$what"}} if ( exists $self->{"_$what"} ) ;
	return ();
}

#
# Given a baseline and a pvob (without slashes) search CQ for the record.
# use get("field") to get the values
# It will return an array;
# my ($act) = $bl->get("Activities");
#
package get_baseline_record_details;
 
sub new { 
	my $class = shift;
	my $self = {@_};
	bless($self, $class);
	my $results = $self->get_baseline_record_details(@_);
	return $self if $results;
	return undef;
}

sub get_baseline_record_details {
	my $self = shift;
	my %lopts = @_;
	my $queryDef = $lopts{session}->BuildQuery("ALMBaseline");
	# TODO figure out how to do real "top" top=>1

	# Get all the columns for now.
	# my $failure = $queryDef->BuildField("Name");
	my $entdef = $lopts{session}->GetEntityDef("ALMBaseline");
	my $names = $entdef->GetFieldDefNames();
	
	for my $name (@$names ) {
	   my $failure = $queryDef->BuildField($name);
	}
	  
	my $operator = $queryDef->CreateTopNode($CQPerlExt::CQ_BOOL_OP_AND); 
	$operator->BuildFilter("PVOB_OrLocation", $CQPerlExt::CQ_COMP_OP_EQ, ["$lopts{pvob}"]);
	$operator->BuildFilter("Name", $CQPerlExt::CQ_COMP_OP_EQ, ["$lopts{baseline}"]);

	my $queryfielddefs = $queryDef->GetQueryFieldDefs(); 
	#my $baselinefield = $queryfielddefs->ItemByName("history.action_timestamp");
	#$baselinefield->SetSortType($CQPerlExt::CQ_SORT_DESC); 
	#$baselinefield->SetSortOrder(1);

	my $resultSet;
	eval {$resultSet = $lopts{session}->BuildResultSet($queryDef)};
	# Write to log file
	if ($@) {
		$logger->log_print_error(__LINE__ . "NG[̃rhɎs܂: $@"); 
		return undef;
	}
	else {
		#$logger->log_print_message(" f[^x[XNG[ʂ𐶐Ă܂");
		1;
	}

	my %results;

	my $ct = $resultSet->ExecuteAndCountRecords(); 
	# Find the first record within ClearQuest
	return undef if ($ct == 0);

	OUTER: for (my $i = 0; $i < $ct; $i++){
	$resultSet->MoveNext;
		for (my $i = 1; $i <= $resultSet->GetNumberOfColumns(); $i++) {
			my $name = $resultSet->GetColumnLabel($i);
			#my $type = $resultSet->GetColumnType($i);
			my $value = $resultSet->GetColumnValue($i);
			$self->{$name}{$value}++;
			last OUTER if $opts{top};
			
		}
	}
	#CQPerlExt::CQSession_Unbuild($lopts{session});
	1;
}	

# Always returns an array - TODO should add "wantarray"
sub get {
	my $self = shift;
	my $what = shift;
	my @values;
	#TODO "Tie" this
	if (exists $self->{$what}) {
		if (ref $self->{$what} eq "HASH") {
			for my $k (keys %{$self->{$what}}){
				push @values, $k;
			}
		}
		elsif( ref $self->{$what} eq "ARRAY"){
			for my $k (@{$self->{$what}}) {
				push @values, $k;
			}
		}
		else {
			push @values,$self->{$what};
		}
		return @values;
	}
	else {
		return undef;
	}
} 

#### End of script
